class PaperDimensions:
    def __init__(
        self,
        columns: int,
        column_width: float,
        text_width: float,
        font_size: int = 10,
        paper_width: float = 8.5,
    ):
        self.columns = columns
        self.column_width = column_width
        self.text_width = text_width
        self.font_size = font_size
        self.paper_width = paper_width


ICML_Dimensions = PaperDimensions(
    columns=2,
    column_width=3.25,
    text_width=6.75,
)

NeurIPS_Dimensions = PaperDimensions(
    columns=1,
    column_width=5.5,
    text_width=5.5,
)

AAAI_Dimensions = PaperDimensions(
    columns=2,
    column_width=3.3125, # TODO: confirm if this is correct
    text_width=7.0,
)

JMLR_Dimensions = PaperDimensions(
    columns=1,
    column_width=6.0,
    text_width=6.0,
)

def getDefaultDimensions(venue: str):
    venue = venue.lower()

    if venue == 'icml': return ICML_Dimensions
    if venue == 'neurips': return NeurIPS_Dimensions
    if venue == 'aaai': return AAAI_Dimensions
    if venue == 'jmlr': return JMLR_Dimensions

    raise NotImplementedError(f"Sorry, don't know this conference: {venue}")



import numpy as np
import pandas as pd
from matplotlib.axes import Axes
from dataclasses import dataclass
from typing import Any, Dict, List
from PyExpUtils.results.tools import subsetDF, splitByValue
from PyExpPlotting.colors import ColorPalette
from PyExpPlotting.tools import reduceCurve, CurveReducer, getBest

# TODO: some of these are true of all plots
# could set up an inheritance system
@dataclass
class SensitivityOptions:
    hypers: List[str] | None = None
    style: str = 'solid'
    color: ColorPalette = ColorPalette()
    label: str | None = None
    width: float = 1.0
    height: float = 1.0
    legend: bool = True
    linewidth: float = 1.0
    log_x: int | None = None
    x_label: str | None = None
    y_label: str | None = None

    prefer: str = 'big'
    curve_reducer: str | CurveReducer = 'auc'
    param_reducer: str = 'best'

def _getHypers(df: pd.DataFrame, param: str, metric: str):
    return set(df.columns) - {param, metric}

def sliceBestPerformance(df: pd.DataFrame, param: str, metric: str, options: SensitivityOptions):
    perf = reduceCurve(df, metric, options.curve_reducer)
    idx = getBest(perf, options.prefer)
    row = df.iloc[idx]

    hypers = options.hypers or _getHypers(df, param, metric)
    conds = { k: row[k] for k in hypers }
    parts = subsetDF(df, conds)

    param_vals = df[param].unique()
    param_vals.sort()

    out: Dict[Any, float] = {}
    for p in param_vals:
        sub = parts[parts[param] == p]
        out[p] = sub[metric]

    return out

def bestPerformanceEach(df: pd.DataFrame, param: str, metric: str, options: SensitivityOptions):
    df = df.copy()
    df['_perf'] = reduceCurve(df, metric, options.curve_reducer)

    out: Dict[Any, float] = {}
    for (p, sub) in splitByValue(df, param):
        idx = getBest(sub['_perf'], options.prefer)
        row = sub.iloc[idx]
        out[p] = row['_perf']

    return out

def plotSensitivityCurve(df: pd.DataFrame, ax: Axes, param: str, metric: str, options: SensitivityOptions):
    if options.param_reducer == 'slice':
        parts = sliceBestPerformance(df, param, metric, options)

    elif options.param_reducer == 'best':
        parts = bestPerformanceEach(df, param, metric, options)

    else:
        raise Exception('Unknown parameter reduction strategy')

    x = np.asarray(parts.keys())
    y = np.asarray(parts.values())

    idxs = np.argsort(x)
    x = x[idxs]
    y = y[idxs]

    ax.plot(
        x, y,
        label=options.label,
        color=options.color.get(options.label),
        linewidth=options.linewidth,
    )

    # TODO: make this a shared method
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)

    if options.log_x is not None:
        ax.set_xscale('log', base=options.log_x)

    if options.legend:
        ax.legend(frameon=False)

    if options.x_label is None:
        x_label = param.split('.')[-1]
        if options.log_x is not None:
            scale = options.log_x
            x_label += f' (log-{scale} scale)'

        ax.set_xlabel(x_label)

    elif options.x_label:
        ax.set_xlabel(options.x_label)



import os
import sys
import glob
import numpy as np

sys.path.append(os.getcwd())

import matplotlib.pyplot as plt
from PyExpUtils.models.ExperimentDescription import loadExperiment, ExperimentDescription
from PyExpUtils.results.results import loadResults, getBest, splitOverParameter

from PyExpPlotting.matplot import save, setDefaultConference
from PyExpPlotting.distributions import parameterStudyAlgs, plot, stacked, stackedAlgs

setDefaultConference('icml')

colors = {
    'QLearning': 'blue',
    'SBEED': 'black',
    'QRC': 'red',
}

class Model(ExperimentDescription):
    def __init__(self, d, path):
        super().__init__(d, path, save_key='tests/results/{name}/{agent}')
        self.agent = d['agent']
        self.problem = d['problem']

if __name__ == '__main__':
    f, ax = plt.subplots(1)

    alg_results = []
    exp_paths = glob.glob('tests/experiments/example/CartPole/*.json')
    for exp_path in exp_paths:
        exp = loadExperiment(exp_path, Model)

        results = loadResults(exp, 'returns.csv')
        alg_results.append(results)

    parameterStudyAlgs(alg_results, ax, 'optimizer.alpha', np.mean, {
        'colors': colors,
        'prefer': 'big',
        'alg_spacing': 0.25,
        'hist': True,
        'process_y_ticks': np.log2,
        'highlight_top': True,
    })

    save('tests/', 'test', height_ratio=2.0)



import os
import sys

sys.path.append(os.getcwd())

import matplotlib.pyplot as plt
from PyExpUtils.models.ExperimentDescription import loadExperiment, ExperimentDescription
from PyExpUtils.results.results import loadResults

from PyExpPlotting.tools import findExperiments
from PyExpPlotting.matplot import save, setDefaultConference
from PyExpPlotting.learning_curves import plotBest

setDefaultConference('icml')

class Model(ExperimentDescription):
    def __init__(self, d, path):
        super().__init__(d, path, save_key='tests/results/{name}/{agent}')
        self.agent = d['agent']
        self.problem = d['problem']

if __name__ == '__main__':

    exps = findExperiments(key='{domain}')
    for domain in exps:
        print(domain)
        f, ax = plt.subplots(1)
        for exp_path in exps[domain]:
            exp = loadExperiment(exp_path, Model)

            results = loadResults(exp, 'returns.csv')

            plotBest(results, ax)

        plt.legend()
        save('tests/', 'test', height_ratio=0.75, f=f)



import numpy as np
from typing import Any, Callable, Dict, List, Optional
from PyExpUtils.results.backends.backend import DuckResult, ResultList
from PyExpUtils.results.results import splitOverParameter, getBest
from scipy.stats.kde import gaussian_kde
import PyExpUtils.utils.dict as DictUtils

# TODO: split this into 2 dictionaries. One that contains user-specified options
# and another that contains context information (e.g. best params for an algorithm)
# also, when we make the split, organize the options into a meaningful order somehow

# TODO: see if we can use python 3.8+ and use typed-dicts instead, that way we can
# start type-checking this library
def buildOptions(options: Optional[Dict[str, Any]]):
    options = options if options is not None else {}
    out_options = {}

    out_options['prefer'] = options.get('prefer', 'big')
    out_options['y_offset'] = options.get('y_offset', 0)
    out_options['spacing'] = options.get('spacing', 1.0)
    out_options['alg_spacing'] = options.get('alg_spacing', 0.75)
    out_options['x_range'] = options.get('x_range')
    out_options['color'] = options.get('color')
    out_options['colors'] = options.get('colors', {})
    out_options['label'] = options.get('label')
    out_options['bins'] = options.get('bins', 30)
    out_options['alpha'] = options.get('alpha', 0.6)
    out_options['fidelity'] = options.get('fidelity', 200)
    out_options['kde'] = options.get('kde', True)
    out_options['hist'] = options.get('hist', True)
    out_options['dist_height'] = options.get('dist_height', 0.9)
    out_options['y_ticks'] = options.get('y_ticks', True)
    out_options['process_y_ticks'] = options.get('process_y_ticks', lambda x: x)
    out_options['highlight_top'] = options.get('highlight_top', False)
    out_options['highlight_level'] = options.get('highlight_level', 0.2)
    out_options['highlight'] = options.get('highlight', True)
    out_options['best_params'] = options.get('best_params')  # <-- actually a context variable
    out_options['best_param'] = options.get('best_param')  # <-- also context, and worse only useful for some of the funcs but not all

    return out_options

def minMaxScale(x):
    return (x - np.min(x)) / (np.max(x) - np.min(x))

def pullLeft(x, p):
    if x < 0:
        return (1 + p) * x

    return (1 - p) * x

def pullRight(x, p):
    return -pullLeft(-x, p)

def plot(result: DuckResult, ax, reducer: Callable[[np.ndarray], float], options: Optional[Dict[str, Any]] = None):
    o = buildOptions(options)

    if not o['hist'] and not o['kde']:
        raise Exception("Well if you don't want a histogram or kde, then what *do* you want??")

    curves = result.load()
    points = [reducer(curve) for curve in curves]

    if o['x_range'] is None:
        lo = np.min(points)
        hi = np.max(points)

    else:
        lo, hi = o['x_range']

    lo = pullLeft(lo, 0.02)
    hi = pullRight(hi, 0.02)

    alpha = 1
    if not o['highlight']:
        alpha = o['highlight_level']

    if o['hist']:
        bins = o['bins']
        hist, _ = np.histogram(points, bins=bins, range=(lo, hi))
        hist = minMaxScale(hist) * o['dist_height']
        hist_x = np.linspace(lo, hi, bins)

        ax.bar(hist_x, hist, bottom=o['y_offset'], width=(hi - lo) / bins, color=o['color'], alpha=o['alpha'] * alpha)

    if o['kde']:
        kde = gaussian_kde(points)
        dist_space = np.linspace(lo, hi, o['fidelity'])
        dist = minMaxScale(kde(dist_space)) * o['dist_height'] + o['y_offset']
        ax.plot(dist_space, dist, linewidth=1.0, color=o['color'], alpha=alpha)

        # if we only want a KDE, then shade in the region under the curve
        if not o['hist']:
            ax.fill_between(dist_space, np.zeros(o['fidelity']) + o['y_offset'], dist, color=o['color'], alpha=0.2 * alpha)

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)

def stacked(result_dict: Dict[Any, DuckResult], ax, reducer: Callable[[np.ndarray], float], options: Optional[Dict[str, Any]] = None):
    o = buildOptions(options)

    y_ticks = []
    y_offset = o['y_offset']
    for key in result_dict:
        o['y_offset'] = y_offset

        if o['highlight_top'] and o['best_param'] == key:
            o['highlight'] = True
        elif o['highlight_top']:
            o['highlight'] = False

        plot(result_dict[key], ax, reducer, o)

        y_ticks.append(y_offset)
        y_offset += o['spacing']

    empty_ticks = all([tick.get_text() == '' for tick in ax.get_yticklabels()])
    if empty_ticks and o['y_ticks']:
        y_labels = list(result_dict)
        ax.yaxis.set_ticks(y_ticks)
        y_labels = list(map(o['process_y_ticks'], y_labels))
        ax.set_yticklabels(y_labels)

def stackedAlgs(result_dicts: List[Dict[Any, DuckResult]], ax, reducer: Callable[[np.ndarray], float], options: Optional[Dict[str, Any]] = None):
    o = buildOptions(options)

    lo = np.inf
    hi = -np.inf

    for alg_dict in result_dicts:
        for key in alg_dict:
            curves = alg_dict[key].load()
            points = [reducer(curve) for curve in curves]

            lo = np.min(points + [lo])
            hi = np.max(points + [hi])

    o['x_range'] = (lo, hi)
    y_offset = 0
    center_alg = int(len(result_dicts) // 2)
    for idx, alg_dict in enumerate(result_dicts):
        o['y_offset'] = y_offset
        o['y_ticks'] = idx == center_alg

        key = list(alg_dict)[0]

        if hasattr(alg_dict[key].exp, 'agent'):
            alg = alg_dict[key].exp.agent
            o['color'] = o['colors'].get(alg)

        if o['best_params']:
            o['best_param'] = o['best_params'][idx]

        stacked(alg_dict, ax, reducer, o)

        y_offset += o['alg_spacing']

def parameterStudyAlgs(
        results: List[ResultList],
        ax,
        param: str,
        reducer: Callable[[np.ndarray], float],
        options: Optional[Dict[str, Any]] = None,
    ):
    o = buildOptions(options)

    # account for the fact that there are multiple algorithms
    o['spacing'] = len(results) * o['alg_spacing'] + o['spacing']

    alg_dicts = []
    best_params = []
    for alg_results in results:
        alg_results = list(alg_results)
        total_best = getBest(alg_results, prefer=o['prefer'])
        best_params.append(DictUtils.get(total_best.params, param))

        param_dict = splitOverParameter(alg_results, param)

        result_dict = {}
        for key in param_dict:
            result_dict[key] = getBest(param_dict[key], prefer=o['prefer'])

        alg_dicts.append(result_dict)

    o['best_params'] = best_params
    stackedAlgs(alg_dicts, ax, reducer, o)



import numpy as np
import pandas as pd
from typing import Callable

CurveReducer = Callable[[np.ndarray], float]

def tail_mean(m: np.ndarray):
    l = int(m.shape[0] * 0.1)
    return float(np.mean(m[l:]))

def getCurveReducer(reducer: str | CurveReducer) -> CurveReducer:
    if reducer == 'auc':
        return np.mean

    if reducer == 'end':
        return tail_mean

    if isinstance(reducer, str):
        raise Exception("Unknown reducer type")

    return reducer

def reduceCurve(df: pd.DataFrame, col: str, reducer: str | CurveReducer):
    cr = getCurveReducer(reducer)
    d = df[col].apply(cr)
    return np.asarray(d)

def getBest(data: np.ndarray, prefer: str):
    if prefer == 'big':
        return data.argmax()

    elif prefer == 'small':
        return data.argmin()

    raise Exception('Unknown getBest preference')



from typing import Dict

class ColorPalette:
    def __init__(self):
        self._mapper: Dict[str, str] = {}
        self._idx = 0

    def get(self, label: str | None):
        if label is None:
            label = ''

        if label not in self._mapper:
            self._mapper[label] = self.next()

        return self._mapper[label]

    def lock(self, colors: Dict[str, str]):
        self._mapper |= colors
        return self

    def next(self):
        self._idx += 1
        return ''



import numpy as np
from dataclasses import dataclass
from matplotlib.axes import Axes

from PyExpPlotting.colors import ColorPalette

@dataclass
class LineplotOptions:
    label: str | None = None
    color: ColorPalette = ColorPalette()

    linewidth: float = 0.25
    alpha: float = 0.2

    legend: bool = True
    x_label: str | None = None
    y_label: str | None = None
    title: str | None = None

def plotLearningCurve(data: np.ndarray, ax: Axes, lo: np.ndarray | None = None, hi: np.ndarray | None = None, options: LineplotOptions = LineplotOptions()):
    color = options.color.get(options.label)
    ax.plot(data, label=options.label, color=color, linewidth=options.linewidth)

    if lo is not None and hi is not None:
        ax.fill_between(range(data.shape[0]), lo, hi, color=color, alpha=options.alpha)

    # TODO: make this a shared method
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)

    if options.legend:
        ax.legend(frameon=False)

    if options.x_label is not None:
        ax.set_xlabel(options.x_label)

    if options.y_label is not None:
        ax.set_ylabel(options.y_label)

    if options.title is not None:
        ax.set_title(options.title)

@dataclass
class AllLinesPlotOptions(LineplotOptions):
    summary: str | None = 'mean'
    max_lines: int = 30

def plotAllLearningCurves(data: np.ndarray, ax: Axes, options: AllLinesPlotOptions = AllLinesPlotOptions()):
    color = options.color.get(options.label)

    assert data.ndim == 2
    idxs = np.arange(data.shape[0])
    if options.max_lines < data.shape[0]:
        top = int(options.max_lines / 2)
        bottom = options.max_lines - top

        scores = np.mean(data, axis=1)
        sorted_idxs = np.argsort(scores)
        bottom_idxs = sorted_idxs[:bottom]
        top_idxs = sorted_idxs[-top:]

        idxs = np.concatenate((bottom_idxs, top_idxs))

    label = options.label
    for idx in idxs:
        ax.plot(data[idx], label=label, color=color, linewidth=options.linewidth, alpha=options.alpha)
        label = None

    if options.summary == 'mean':
        mean = np.mean(data, axis=0)
        ax.plot(mean, color=color, linewidth=options.linewidth)

    # TODO: make this a shared method
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)

    if options.legend:
        ax.legend(frameon=False)

    if options.x_label is not None:
        ax.set_xlabel(options.x_label)

    if options.y_label is not None:
        ax.set_ylabel(options.y_label)

    if options.title is not None:
        ax.set_title(options.title)



import os
import sys
import glob

sys.path.append(os.getcwd())

import matplotlib.pyplot as plt
from PyExpUtils.models.ExperimentDescription import loadExperiment, ExperimentDescription
from PyExpUtils.results.results import loadResults

from PyExpPlotting.matplot import save, setDefaultConference
from PyExpPlotting.learning_curves import plotBest

setDefaultConference('icml')

class Model(ExperimentDescription):
    def __init__(self, d, path):
        super().__init__(d, path, save_key='tests/results/{name}/{agent}')
        self.agent = d['agent']
        self.problem = d['problem']

if __name__ == '__main__':
    f, ax = plt.subplots(1)

    exp_paths = glob.glob('tests/experiments/example/CartPole/*.json')
    for exp_path in exp_paths:
        exp = loadExperiment(exp_path, Model)

        results = loadResults(exp, 'returns.csv')

        plotBest(results, ax, {
            'y_label': 'return',
            'x_label': 'steps',
        })

    save('tests/', 'test', height_ratio=0.75)



from typing import Optional, Union
import os
import importlib
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
from PyExpPlotting.defaults import ICML_Dimensions, PaperDimensions, getDefaultDimensions

default_conference = ICML_Dimensions
def setDefaultConference(conference: Union[str, PaperDimensions]):
    global default_conference
    if type(conference) is str:
        default_conference = getDefaultDimensions(conference)

    elif type(conference) is PaperDimensions:
        default_conference = conference

    else:
        raise NotImplementedError('Should be unreachable, but types say otherwise')

    setFonts(default_conference.font_size)

def setFonts(font_size: int):
    small = font_size - 4
    medium = font_size - 2
    large = font_size

    plt.rc('font', size=small)          # controls default text sizes
    plt.rc('axes', titlesize=medium)    # fontsize of the axes title
    plt.rc('axes', labelsize=medium)    # fontsize of the x and y labels
    plt.rc('xtick', labelsize=small)    # fontsize of the tick labels
    plt.rc('ytick', labelsize=small)    # fontsize of the tick labels
    plt.rc('legend', fontsize=medium)   # legend fontsize
    plt.rc('figure', titlesize=large)   # fontsize of the figure title

def save(
        save_path: str,
        plot_name: str,
        plot_type: Optional[str] = None,
        dims: Optional[PaperDimensions] = None,
        save_type: str = 'png',
        width: float = 1.0,
        height_ratio: float = 1.0,
        f: Optional[Figure] = None,
    ):

    # don't make this a default because it could be changed
    # after this function is parsed
    if f is None:
        f = plt.gcf()

    # likewise could be changed after function parse
    # so cannot be a default parameter
    if not dims:
        dims = default_conference

    # too much logic to stick in default parameters
    if plot_type is None:
        main_file = importlib.import_module('__main__').__file__
        assert main_file is not None
        plot_type = os.path.basename(main_file).replace('.py', '').replace('_', '-')

    save_path = f'{save_path}/{plot_type}'
    os.makedirs(save_path, exist_ok=True)

    width = width * dims.column_width
    height = width * height_ratio
    f.set_size_inches((width, height), forward=True)
    f.savefig(f'{save_path}/{plot_name}.{save_type}', dpi=600, bbox_inches='tight')
